@@ -1,6 +1,5 @@
 import asyncio
 import contextvars
-import gc
 import inspect
 import logging
 import warnings
@@ -9,7 +8,7 @@ from collections import defaultdict, deque
 from collections.abc import Callable
 from contextvars import ContextVar
 from pathlib import Path
-from typing import Any, Literal, TypeVar, overload
+from typing import Any, Literal, TypeVar, cast, overload
 
 import anyio  # pyright: ignore[reportMissingImports]
 from uuid_extensions import uuid7str  # pyright: ignore[reportMissingImports, reportUnknownVariableType]
@@ -30,6 +29,7 @@ from bubus.models import (
     PythonIdentifierStr,
     PythonIdStr,
     T_Event,
+    T_EventResultType,
     UUIDStr,
     get_handler_id,
     get_handler_name,
@@ -46,9 +46,10 @@ class QueueShutDown(Exception):
     pass
 
 
-QueueEntryType = TypeVar('QueueEntryType', bound=BaseEvent)
+QueueEntryType = TypeVar('QueueEntryType', bound='BaseEvent[Any]')
+T_ExpectedEvent = TypeVar('T_ExpectedEvent', bound='BaseEvent[Any]')
 
-EventPatternType = PythonIdentifierStr | Literal['*'] | type[BaseEvent]
+EventPatternType = PythonIdentifierStr | Literal['*'] | type['BaseEvent[Any]']
 
 
 class CleanShutdownQueue(asyncio.Queue[QueueEntryType]):
@@ -133,11 +134,13 @@ class CleanShutdownQueue(asyncio.Queue[QueueEntryType]):
 
 
 # Context variable to track the current event being processed (for setting event_parent_id from inside a child event)
-_current_event_context: ContextVar[BaseEvent | None] = ContextVar('current_event', default=None)
+_current_event_context: ContextVar['BaseEvent[Any] | None'] = ContextVar('current_event', default=None)
 # Context variable to track if we're inside a handler (for nested event detection)
 inside_handler_context: ContextVar[bool] = ContextVar('inside_handler', default=False)
 # Context variable to track if we hold the global lock (for re-entrancy across tasks)
 holds_global_lock: ContextVar[bool] = ContextVar('holds_global_lock', default=False)
+# Context variable to track the current handler ID (for tracking child events)
+_current_handler_id_context: ContextVar[str | None] = ContextVar('current_handler_id', default=None)
 
 
 class ReentrantLock:
@@ -247,9 +250,9 @@ class EventBus:
 
     # Runtime State
     id: UUIDStr = '00000000-0000-0000-0000-000000000000'
-    handlers: dict[PythonIdStr, list[ContravariantEventHandler[BaseEvent]]]  # collected by .on(<event_type>, <handler>)
-    event_queue: CleanShutdownQueue[BaseEvent] | None
-    event_history: dict[UUIDStr, BaseEvent]  # collected by .dispatch(<event>)
+    handlers: dict[PythonIdStr, list[ContravariantEventHandler['BaseEvent[Any]']]]  # collected by .on(<event_type>, <handler>)
+    event_queue: CleanShutdownQueue['BaseEvent[Any]'] | None
+    event_history: dict[UUIDStr, 'BaseEvent[Any]']  # collected by .dispatch(<event>)
 
     _is_running: bool = False
     _runloop_task: asyncio.Task[None] | None = None
@@ -267,14 +270,46 @@ class EventBus:
         assert self.name.isidentifier(), f'EventBus name must be a unique identifier string, got: {self.name}'
 
         # Force garbage collection to clean up any dead EventBus instances in the WeakSet
-        gc.collect()
+        # gc.collect()  # Commented out - this is expensive and causes 5s delays when creating many EventBus instances
 
         # Check for name uniqueness among existing instances
-        for existing_bus in EventBus.all_instances:
+        # We'll collect potential conflicts and check if they're still alive
+        original_name = self.name
+        conflicting_buses: list[EventBus] = []
+
+        for existing_bus in list(EventBus.all_instances):  # Make a list copy to avoid modification during iteration
             if existing_bus is not self and existing_bus.name == self.name:
-                raise ValueError(
-                    f'EventBus with name "{self.name}" already exists. Please choose a unique name or let it auto-generate.'
-                )
+                # Try to trigger collection of just this object by checking if it's collectable
+                # First, temporarily remove from WeakSet to see if that was the only reference
+                EventBus.all_instances.discard(existing_bus)
+
+                # Check if the object is still reachable by creating a new weak reference
+                # If the object only existed in the WeakSet, it should be unreachable now
+                try:
+                    # Try to access an attribute to see if the object is still valid
+                    _ = existing_bus.name  # This will work if object is still alive
+
+                    # Object is still alive with real references, restore to WeakSet
+                    EventBus.all_instances.add(existing_bus)
+                    conflicting_buses.append(existing_bus)
+                except Exception:
+                    # Object was garbage collected or is invalid (e.g., AttributeError), that's fine
+                    # Don't re-add to WeakSet, let it stay removed
+                    pass
+
+        # If we found conflicting buses, auto-generate a unique suffix
+        if conflicting_buses:
+            # Generate a unique suffix using the last 8 chars of a UUID
+            unique_suffix = uuid7str()[-8:]
+            self.name = f'{original_name}_{unique_suffix}'
+
+            warnings.warn(
+                f'⚠️ EventBus with name "{original_name}" already exists. '
+                f'Auto-generated unique name: "{self.name}" to avoid conflicts. '
+                f'Consider using unique names or stop(clear=True) on unused buses.',
+                UserWarning,
+                stacklevel=2,
+            )
 
         self.event_queue = None
         self.event_history = {}
@@ -320,19 +355,19 @@ class EventBus:
         return str(self)
 
     @property
-    def events_pending(self) -> list[BaseEvent]:
+    def events_pending(self) -> list['BaseEvent[Any]']:
         """Get events that haven't started processing yet (does not include events that have not even finished dispatching yet in self.event_queue)"""
         return [
             event for event in self.event_history.values() if event.event_started_at is None and event.event_completed_at is None
         ]
 
     @property
-    def events_started(self) -> list[BaseEvent]:
+    def events_started(self) -> list['BaseEvent[Any]']:
         """Get events currently being processed"""
         return [event for event in self.event_history.values() if event.event_started_at and not event.event_completed_at]
 
     @property
-    def events_completed(self) -> list[BaseEvent]:
+    def events_completed(self) -> list['BaseEvent[Any]']:
         """Get events that have completed processing"""
         return [event for event in self.event_history.values() if event.event_completed_at is not None]
 
@@ -357,15 +392,15 @@ class EventBus:
 
     # 5. EventHandlerClassMethod[BaseEvent] - sync classmethod taking cls and event
     @overload
-    def on(self, event_pattern: EventPatternType, handler: EventHandlerClassMethod[BaseEvent]) -> None: ...
+    def on(self, event_pattern: EventPatternType, handler: EventHandlerClassMethod['BaseEvent[Any]']) -> None: ...
 
     # 6. AsyncEventHandlerClassMethod[BaseEvent] - async classmethod taking cls and event
     @overload
-    def on(self, event_pattern: EventPatternType, handler: AsyncEventHandlerClassMethod[BaseEvent]) -> None: ...
+    def on(self, event_pattern: EventPatternType, handler: AsyncEventHandlerClassMethod['BaseEvent[Any]']) -> None: ...
 
     # I dont think this is needed, but leaving it here for now
     # 9. Coroutine[Any, Any, Any] - direct coroutine
-    # @overload
+    # @overload # type: ignore[reportUnknownReturnType]
     # def on(self, event_pattern: EventPatternType, handler: Coroutine[Any, Any, Any]) -> None: ...
 
     def on(
@@ -373,11 +408,11 @@ class EventBus:
         event_pattern: EventPatternType,
         handler: (  # TypeAlias with args doesnt work on overloaded signature, has to be defined inline
             EventHandlerFunc[T_Event]
-            | AsyncEventHandlerFunc[BaseEvent]
+            | AsyncEventHandlerFunc['BaseEvent[Any]']
             | EventHandlerMethod[T_Event]
-            | AsyncEventHandlerMethod[BaseEvent]
-            | EventHandlerClassMethod[BaseEvent]
-            | AsyncEventHandlerClassMethod[BaseEvent]
+            | AsyncEventHandlerMethod['BaseEvent[Any]']
+            | EventHandlerClassMethod['BaseEvent[Any]']
+            | AsyncEventHandlerClassMethod['BaseEvent[Any]']
         ),
     ) -> None:
         """
@@ -430,7 +465,7 @@ class EventBus:
         self.handlers[event_key].append(handler)  # type: ignore
         logger.debug(f'👂 {self}.on({event_key}, {get_handler_name(handler)}) Registered event handler')
 
-    def dispatch(self, event: BaseEvent) -> BaseEvent:
+    def dispatch(self, event: 'BaseEvent[Any]') -> 'BaseEvent[Any]':
         """
         Enqueue an event for processing and immediately return an Event(status='pending') version (synchronous).
         You can then await the Event(status='pending') object to block until its Event(status='completed') versionis ready,
@@ -461,10 +496,20 @@ class EventBus:
 
         # Automatically set event_parent_id from context if not already set
         if event.event_parent_id is None:
-            current_event = _current_event_context.get()
+            current_event: 'BaseEvent[Any] | None' = _current_event_context.get()
             if current_event is not None:
                 event.event_parent_id = current_event.event_id
 
+        # Track child events - if we're inside a handler, add this event to the handler's event_children list
+        # Only track if this is a NEW event (not forwarding an existing event)
+        current_handler_id = _current_handler_id_context.get()
+        if current_handler_id is not None and inside_handler_context.get():
+            current_event = _current_event_context.get()
+            if current_event is not None and current_handler_id in current_event.event_results:
+                # Only add as child if it's a different event (not forwarding the same event)
+                if event.event_id != current_event.event_id:
+                    current_event.event_results[current_handler_id].event_children.append(event)
+
         # Add this EventBus to the event_path if not already there
         if self.name not in event.event_path:
             # preserve identity of the original object instead of creating a new one, so that the original object remains awaitable to get the result
@@ -525,19 +570,43 @@ class EventBus:
 
         return event
 
+    @overload
     async def expect(
         self,
-        event_type: PythonIdentifierStr | type[BaseEvent],
+        event_type: type[T_ExpectedEvent],
+        include: Callable[['BaseEvent[Any]' | T_ExpectedEvent], bool] = lambda _: True,
+        exclude: Callable[['BaseEvent[Any]' | T_ExpectedEvent], bool] = lambda _: False,
+        predicate: Callable[['BaseEvent[Any]' | T_ExpectedEvent], bool] = lambda _: True,  # deprecated, alias for include
         timeout: float | None = None,
-        predicate: Callable[[BaseEvent], bool] | None = None,
-    ) -> BaseEvent:
+    ) -> T_ExpectedEvent: ...
+
+    @overload
+    async def expect(
+        self,
+        event_type: PythonIdentifierStr,
+        include: Callable[['BaseEvent[Any]'], bool] = lambda _: True,
+        exclude: Callable[['BaseEvent[Any]'], bool] = lambda _: False,
+        predicate: Callable[['BaseEvent[Any]'], bool] = lambda _: True,  # deprecated, alias for include
+        timeout: float | None = None,
+    ) -> 'BaseEvent[Any]': ...
+
+    async def expect(
+        self,
+        event_type: PythonIdentifierStr | type[T_ExpectedEvent],
+        include: Callable[['BaseEvent[Any]'], bool] = lambda _: True,
+        exclude: Callable[['BaseEvent[Any]'], bool] = lambda _: False,
+        predicate: Callable[['BaseEvent[Any]'], bool] = lambda _: True,  # deprecated, alias for include
+        timeout: float | None = None,
+    ) -> 'BaseEvent[Any]' | T_ExpectedEvent:
         """
-        Wait for an event matching the given type/pattern with optional predicate filter.
+        Wait for an event matching the given type/pattern with optional filters.
 
         Args:
                 event_type: The event type string or model class to wait for
-                timeout: Maximum time to wait in seconds (None = wait forever)
-                predicate: Optional filter function that must return True for the event to match
+                include: Filter function that must return True for the event to match (default: lambda e: True)
+                exclude: Filter function that must return False for the event to match (default: lambda e: False)
+                predicate: Deprecated name, alias for include (default: lambda e: True)
+                timeout: Maximum time to wait in seconds as a float (None = wait forever)
 
         Returns:
                 The first matching event
@@ -549,25 +618,38 @@ class EventBus:
                 # Wait for any response event
                 response = await eventbus.expect('ResponseEvent', timeout=30)
 
-                # Wait for specific response with predicate
+                # Wait for specific response with include filter
                 response = await eventbus.expect(
                         'ResponseEvent',
-                        predicate=lambda e: e.request_id == my_request_id,
+                        include=lambda e: e.request_id == my_request_id,
+                        timeout=30
+                )
+
+                # Wait for response excluding certain types
+                response = await eventbus.expect(
+                        'ResponseEvent',
+                        exclude=lambda e: e.error_code is not None,
                         timeout=30
                 )
         """
-        future: asyncio.Future[BaseEvent] = asyncio.Future()
+        future: asyncio.Future['BaseEvent[Any]'] = asyncio.Future()
 
-        def notify_expect_handler(event: BaseEvent) -> None:
+        # Handle backwards compatibility: merge predicate into include
+        if predicate is not None:  # type: ignore[conditionAlwaysTrue]
+            original_include = include
+            include = lambda e, orig=original_include, pred=predicate: orig(e) and pred(e)
+
+        def notify_expect_handler(event: 'BaseEvent[Any]') -> None:
             """Handler that resolves the future when a matching event is found"""
-            if not future.done() and (predicate is None or predicate(event)):
+            if not future.done() and include(event) and not exclude(event):
                 future.set_result(event)
 
+        # make debugging otherwise ephemeral async expect handlers easier by including some metadata in the stacktrace func names
         current_frame = inspect.currentframe()
         assert current_frame
-        notify_expect_handler.__name__ = f'{self}.expect({event_type}, predicate={predicate and id(predicate)})@{_log_pretty_path(current_frame.f_code.co_filename)}:{current_frame.f_lineno}'  # add file and line number to the name
+        notify_expect_handler.__name__ = f'{self}.expect({event_type}, timeout={timeout})@{_log_pretty_path(current_frame.f_code.co_filename)}:{current_frame.f_lineno}'  # add file and line number to the name
 
-        # Register temporary handler
+        # Register temporary listener that watches for matching events and triggers the expect handler
         self.on(event_type, notify_expect_handler)
 
         try:
@@ -629,7 +711,7 @@ class EventBus:
                 if self.event_queue is None:
                     # Set queue size based on whether we have limits
                     queue_size = 50 if self.max_history_size is not None else 0  # 0 = unlimited
-                    self.event_queue = CleanShutdownQueue[BaseEvent](maxsize=queue_size)
+                    self.event_queue = CleanShutdownQueue['BaseEvent[Any]'](maxsize=queue_size)
                     self._on_idle = asyncio.Event()
                     self._on_idle.clear()  # Start in a busy state unless we confirm queue is empty by running _run_loop_step() at least once
 
@@ -650,8 +732,9 @@ class EventBus:
         if not self._is_running:
             return
 
-        # Wait for completion if timeout specified
-        if timeout:
+        # Wait for completion if timeout specified and > 0
+        # timeout=0 means "don't wait", so skip the wait entirely
+        if timeout is not None and timeout > 0:
             try:
                 await self.wait_until_idle(timeout=timeout)
             except TimeoutError:
@@ -667,21 +750,17 @@ class EventBus:
         # Signal shutdown
         self._is_running = False
 
-        # Wait for the run loop task to finish
-        if self._runloop_task and not self._runloop_task.done():
-            # Give it a short time to finish cleanly
-            done, _pending = await asyncio.wait({self._runloop_task}, timeout=0.1)
+        # Shutdown the queue to unblock any pending get() operations
+        if self.event_queue:
+            self.event_queue.shutdown()
 
-            if not done:
-                # If it doesn't finish in time, cancel it
+        # Wait for the run loop task to finish / force-cancel it if it's hanging
+        if self._runloop_task and not self._runloop_task.done():
+            await asyncio.wait({self._runloop_task}, timeout=0.1)
+            try:
                 self._runloop_task.cancel()
-                try:
-                    await self._runloop_task
-                except asyncio.CancelledError:
-                    pass
-                except Exception:
-                    # logger.debug(f'Exception while stopping {self}: {e}')
-                    pass
+            except Exception:
+                pass
 
         # Clear references
         self._runloop_task = None
@@ -695,6 +774,16 @@ class EventBus:
             # Remove from global instance tracking
             if self in EventBus.all_instances:
                 EventBus.all_instances.discard(self)
+
+            # Remove from event loop's tracking if present
+            try:
+                loop = asyncio.get_running_loop()
+                if hasattr(loop, '_eventbus_instances'):
+                    loop._eventbus_instances.discard(self)  # type: ignore
+            except RuntimeError:
+                # No running loop, that's fine
+                pass
+
             logger.debug(f'🧹 {self} cleared event history and removed from global tracking')
 
         logger.debug(f'🛑 {self} shut down gracefully' if timeout is not None else f'🛑 {self} killed')
@@ -712,17 +801,41 @@ class EventBus:
         self._start()
         assert self._on_idle and self.event_queue, 'EventBus._start() must be called before wait_until_idle() is reached'
 
-        # First wait for the queue to be empty
-        # Then wait for idle state with timeout
-        join_task = asyncio.create_task(self.event_queue.join())
-        idle_task = asyncio.create_task(self._on_idle.wait())
+        start_time = asyncio.get_event_loop().time()
+        remaining_timeout = timeout
 
         try:
-            # Wait for queue to be empty
-            await asyncio.wait_for(join_task, timeout=timeout)
+            # First wait for the queue to be empty
+            join_task = asyncio.create_task(self.event_queue.join())
+            await asyncio.wait_for(join_task, timeout=remaining_timeout)
+
+            # Update remaining timeout
+            if timeout is not None:
+                elapsed = asyncio.get_event_loop().time() - start_time
+                remaining_timeout = max(0, timeout - elapsed)
 
             # Wait for idle state
-            await asyncio.wait_for(idle_task, timeout=timeout)
+            idle_task = asyncio.create_task(self._on_idle.wait())
+            await asyncio.wait_for(idle_task, timeout=remaining_timeout)
+
+            # Critical: Ensure the runloop has settled by yielding control
+            # This allows the runloop to complete any in-flight operations
+            # and prevents race conditions with event_history access
+            await asyncio.sleep(0)  # Yield to event loop
+
+            # Double-check we're truly idle - if new events came in, wait again
+            while not self._on_idle.is_set() or self.events_started or self.events_pending:
+                if timeout is not None:
+                    elapsed = asyncio.get_event_loop().time() - start_time
+                    remaining_timeout = max(0, timeout - elapsed)
+                    if remaining_timeout <= 0:
+                        raise TimeoutError()
+
+                # Clear and wait again
+                self._on_idle.clear()
+                idle_task = asyncio.create_task(self._on_idle.wait())
+                await asyncio.wait_for(idle_task, timeout=remaining_timeout)
+                await asyncio.sleep(0)  # Yield again
 
         except TimeoutError:
             logger.warning(
@@ -735,6 +848,10 @@ class EventBus:
             while self._is_running:
                 try:
                     _processed_event = await self._run_loop_step()
+                    # Check if we should set idle state after processing
+                    if self._on_idle and self.event_queue:
+                        if not (self.events_pending or self.events_started or self.event_queue.qsize()):
+                            self._on_idle.set()
                 except QueueShutDown:
                     # Queue was shut down, exit cleanly
                     break
@@ -756,7 +873,7 @@ class EventBus:
             # Don't call stop() here as it might create new tasks
             self._is_running = False
 
-    async def _get_next_event(self, wait_for_timeout: float = 0.1) -> BaseEvent | None:
+    async def _get_next_event(self, wait_for_timeout: float = 0.1) -> 'BaseEvent[Any] | None':
         """Get the next event from the queue"""
 
         assert self._on_idle and self.event_queue, 'EventBus._start() must be called before _get_next_event()'
@@ -772,6 +889,10 @@ class EventBus:
             # Wait for next event with timeout
             has_next_event, _pending = await asyncio.wait({get_next_queued_event}, timeout=wait_for_timeout)
             if has_next_event:
+                # Check if we're still running before returning the event
+                if not self._is_running:
+                    get_next_queued_event.cancel()
+                    return None
                 return await get_next_queued_event  # await to actually resolve it to the next event
             else:
                 # Get task timed out, cancel it cleanly to suppress warnings
@@ -787,8 +908,8 @@ class EventBus:
             return None
 
     async def _run_loop_step(
-        self, event: BaseEvent | None = None, timeout: float | None = None, wait_for_timeout: float = 0.1
-    ) -> BaseEvent | None:
+        self, event: 'BaseEvent[Any] | None' = None, timeout: float | None = None, wait_for_timeout: float = 0.1
+    ) -> 'BaseEvent[Any] | None':
         """Process a single event from the queue"""
         assert self._on_idle and self.event_queue, 'EventBus._start() must be called before _run_loop_step()'
 
@@ -819,7 +940,7 @@ class EventBus:
         logger.debug(f'✅ {self}._run_loop_step({event}) COMPLETE')
         return event
 
-    async def process_event(self, event: BaseEvent, timeout: float | None = None) -> None:
+    async def process_event(self, event: 'BaseEvent[Any]', timeout: float | None = None) -> None:
         """Process a single event (assumes lock is already held)"""
         # Get applicable handlers
         applicable_handlers = self._get_applicable_handlers(event)
@@ -833,11 +954,36 @@ class EventBus:
         # Mark event as complete if all handlers are done
         event.event_mark_complete_if_all_handlers_completed()
 
+        # After processing this event, check if any parent events can now be marked complete
+        # We do this by walking up the parent chain
+        current = event
+        checked_ids: set[str] = set()
+
+        while current.event_parent_id and current.event_parent_id not in checked_ids:
+            checked_ids.add(current.event_parent_id)
+
+            # Find parent event in any bus's history
+            parent_event = None
+            for bus in EventBus.all_instances:
+                if bus and current.event_parent_id in bus.event_history:
+                    parent_event = bus.event_history[current.event_parent_id]
+                    break
+
+            if not parent_event:
+                break
+
+            # Check if parent can be marked complete
+            if parent_event.event_completed_signal and not parent_event.event_completed_signal.is_set():
+                parent_event.event_mark_complete_if_all_handlers_completed()
+
+            # Move up the chain
+            current = parent_event
+
         # Clean up excess events to prevent memory leaks
         if self.max_history_size:
             self.cleanup_event_history()
 
-    def _get_applicable_handlers(self, event: BaseEvent) -> dict[str, EventHandler]:
+    def _get_applicable_handlers(self, event: 'BaseEvent[Any]') -> dict[str, EventHandler]:
         """Get all handlers that should process the given event, filtering out those that would create loops"""
         applicable_handlers: list[EventHandler] = []
 
@@ -861,7 +1007,7 @@ class EventBus:
         return filtered_handlers
 
     async def _execute_handlers(
-        self, event: BaseEvent, handlers: dict[PythonIdStr, EventHandler] | None = None, timeout: float | None = None
+        self, event: 'BaseEvent[Any]', handlers: dict[PythonIdStr, EventHandler] | None = None, timeout: float | None = None
     ) -> None:
         """Execute all handlers for an event in parallel"""
         applicable_handlers = handlers if (handlers is not None) else self._get_applicable_handlers(event)
@@ -901,7 +1047,9 @@ class EventBus:
                     )
                     pass
 
-    async def _execute_sync_or_async_handler(self, event: BaseEvent, handler: EventHandler, timeout: float | None = None) -> Any:
+    async def _execute_sync_or_async_handler(
+        self, event: 'BaseEvent[T_EventResultType]', handler: EventHandler, timeout: float | None = None
+    ) -> Any:
         """Safely execute a single handler with deadlock detection"""
 
         # Check if this handler has already been executed for this event
@@ -926,6 +1074,8 @@ class EventBus:
         token = _current_event_context.set(event)
         # Mark that we're inside a handler
         handler_token = inside_handler_context.set(True)
+        # Set the current handler ID so child events can be tracked
+        handler_id_token = _current_handler_id_context.set(handler_id)
 
         # Create a task to monitor for potential deadlock / slow handlers
         async def deadlock_monitor():
@@ -962,7 +1112,7 @@ class EventBus:
                 raise ValueError(f'Handler {get_handler_name(handler)} must be a sync or async function, got: {type(handler)}')
 
             logger.debug(
-                f'    ↳ Handler {get_handler_name(handler)}#{handler_id[-4:]} returned: {type(result_value).__name__} {result_value}'  # pyright: ignore[reportUnknownMemberType]
+                f'    ↳ Handler {get_handler_name(handler)}#{handler_id[-4:]} returned: {type(result_value).__name__} {result_value}'  # pyright: ignore
             )
             # Cancel the monitor task since handler completed successfully
             monitor_task.cancel()
@@ -976,7 +1126,7 @@ class EventBus:
                 pass
             else:
                 logger.error(f'    ↳ ERROR: Result not found for {get_handler_name(handler)}#{handler_id[-4:]} after update!')
-            return result_value
+            return cast(T_EventResultType, result_value)
 
         except Exception as e:
             # Cancel the monitor task on error too
@@ -994,6 +1144,7 @@ class EventBus:
             # Reset context
             _current_event_context.reset(token)
             inside_handler_context.reset(handler_token)
+            _current_handler_id_context.reset(handler_id_token)
             # Ensure monitor task is cancelled
             try:
                 if not monitor_task.done():
@@ -1005,7 +1156,7 @@ class EventBus:
                 # logger.debug(f"❌ {self} Handler monitor task cleanup error for {get_handler_name(handler)}#{str(id(handler))[-4:]}({event}): {type(e).__name__}: {e}")
                 pass
 
-    def _would_create_loop(self, event: BaseEvent, handler: EventHandler) -> bool:
+    def _would_create_loop(self, event: 'BaseEvent[Any]', handler: EventHandler) -> bool:
         """Check if calling this handler would create a loop"""
 
         assert inspect.isfunction(handler) or inspect.iscoroutinefunction(handler) or inspect.ismethod(handler), (
@@ -1062,7 +1213,7 @@ class EventBus:
         return False
 
     def _handler_dispatched_ancestor(
-        self, event: BaseEvent, handler_id: str, visited: set[str] | None = None, depth: int = 0
+        self, event: 'BaseEvent[Any]', handler_id: str, visited: set[str] | None = None, depth: int = 0
     ) -> int:
         """Check how many times this handler appears in the ancestry chain. Returns the depth count."""
         # Prevent infinite recursion in case of circular parent references
@@ -1096,14 +1247,14 @@ class EventBus:
         # Recursively check the parent's ancestry
         return self._handler_dispatched_ancestor(parent_event, handler_id, visited, depth)
 
-    async def _default_log_handler(self, event: BaseEvent) -> None:
+    async def _default_log_handler(self, event: 'BaseEvent[Any]') -> None:
         """Default handler that logs all events"""
         # logger.debug(
         # 	f'✅ {self} completed: {event} -> {list(event.event_results.values()) or '<no handlers matched>'}'
         # )
         pass
 
-    async def _default_wal_handler(self, event: BaseEvent) -> None:
+    async def _default_wal_handler(self, event: 'BaseEvent[Any]') -> None:
         """Persist completed event to WAL file as JSONL"""
 
         if not self.wal_path:
@@ -1154,9 +1305,9 @@ class EventBus:
             return 0
 
         # Separate events by status
-        pending_events: list[tuple[str, BaseEvent]] = []
-        started_events: list[tuple[str, BaseEvent]] = []
-        completed_events: list[tuple[str, BaseEvent]] = []
+        pending_events: list[tuple[str, 'BaseEvent[Any]']] = []
+        started_events: list[tuple[str, 'BaseEvent[Any]']] = []
+        completed_events: list[tuple[str, 'BaseEvent[Any]']] = []
 
         for event_id, event in self.event_history.items():
             if event.event_status == 'pending':
